﻿# -*- coding: UTF-8 -*-
#appModules Miranda NG.
# by Marcos Antonio de Oliveira <mao-mail@uol.com.br>.

#This file is covered by the GNU General Public License.

from configobj import ConfigObj
import addonHandler
import api
import speech
import win32con
import wx
import winsound
import time
import tones
from threading import Thread
import gui
import scriptHandler
import ctypes
import os
from keyboardHandler import KeyboardInputGesture
import appModuleHandler
import controlTypes
import winUser
import NVDAObjects.IAccessible
import ui
from time import sleep
import windowUtils
import languageHandler
import displayModel
import queueHandler
import re
import sys

#initialize translation
addonHandler.initTranslation()

#constants
_addonDir = os.path.join(os.path.dirname(__file__), "..").decode("mbcs") # The root of an addon folder
_docFileName = "readme.txt" # The name of an addon documentation file

def clickObject (location, rightMouseButton=False):
	(left, top, width, height) = (int(location[0]), int(location[1]), int(location[2]), int (location[3]))
		# Executa um click com o mouse num objeto
		# Se rightMouseButton = True, clica com o botão direito, caso contrário clica com o esquerdo.
	if (left+top+width+height)>0:
		x=left+(width)/2
		y=top+(height)/2
		winUser.setCursorPos (x,y)
		if rightMouseButton:
			winUser.mouse_event(winUser.MOUSEEVENTF_RIGHTDOWN,0,0,None,None)
			winUser.mouse_event (winUser.MOUSEEVENTF_RIGHTUP,0,0,None,None)
		else:
			winUser.mouse_event(winUser.MOUSEEVENTF_LEFTDOWN,0,0,None,None)
			winUser.mouse_event (winUser.MOUSEEVENTF_LEFTUP,0,0,None,None)

class CheckStatusbarText(NVDAObjects.IAccessible.IAccessible):
	def startMonitoring(self):
		self.isTyping = False
		displayModel.requestTextChangeNotifications(self, True)

	def stopMonitoring(self):
		displayModel.requestTextChangeNotifications(self, False)

	def event_textChange(self):
		queueHandler.queueFunction(queueHandler.eventQueue, self.writing)

	def writing(self):
		# Announces who is typing and start the thread that reproduces the typing sound.
		statusbarText = self.firstChild.name
		if not statusbarText:
			return
		msgIsTyping = _("is typing a message")
		if msgIsTyping in statusbarText:
			if not self.isTyping:
				self.isTyping  = True
				if   AppModule(None).typingAnnounce:
					ui.message (statusbarText[:statusbarText.index (msgIsTyping)+len(msgIsTyping)])
				if AppModule(None).typingSounder:
					th=Thread (target = AppModule(None).playTicks, args = ())
					th.start()
		else:
			self.isTyping  = False


class CheckHistoryChange(NVDAObjects.IAccessible.IAccessible):
	# used to check the change history in windows 10.

	def startMonitoring(self):
		self.lastMessage = None
		displayModel.requestTextChangeNotifications(self, True)

	def stopMonitoring(self):
		displayModel.requestTextChangeNotifications(self, False)

	def event_textChange(self):
		queueHandler.queueFunction(queueHandler.eventQueue, self.change)

	def change (self):
		# Get history content.
		h = self.simpleParent.windowHandle
		historyText = AppModule(None).getHistory(h)
		# Getting the list of messages from the history.
		l = AppModule(None).adjustHistory (historyText, True)
		msg =  (l[-1])
		if self.lastMessage != msg:
			self.lastMessage = msg
			ui.message (msg)

class RadioButtonsFindContact (NVDAObjects.IAccessible.IAccessible):
	def script_focusInRadioButtons (self, gesture):
		key = gesture.mainKeyName
		controlsID = [1393, 1396, 1399]
		id = api.getFocusObject().windowControlID
		index = controlsID.index(id)
		if key in ['downArrow', 'rightArrow']:
			newIndex = index + 1 - 3*(index==2)
		else:
			newIndex = index - 1 + 3*(index==0)
		newID = controlsID[newIndex]
		h = AppModule(None).findDescendantWindow (api.getForegroundObject().windowHandle, newID)
		if h:
			obj = NVDAObjects.IAccessible.getNVDAObjectFromEvent(h, -4, 0)
			obj.setFocus()
			obj.IAccessibleObject.accSelect (2, 0)

	__gestures = {
		"kb:downArrow": "focusInRadioButtons",
		"kb:rightArrow": "focusInRadioButtons",
		"kb:upArrow": "focusInRadioButtons",
		"kb:leftArrow": "focusInRadioButtons",
		}

class AppModule(appModuleHandler.AppModule):
	def __init__(self, *args, **kwargs):
		super(AppModule, self).__init__(*args, **kwargs)
		self.chatWindow = None
		self.statusbar = None
		self._listDialog=None
		self._staticTextDialog=None
		self.dictPositions = {}
		self.lastMessage = None
		self.lastContactMessage = None
		self.lastFocus = None
		self.wmp = None
		self.filterFindContact = 0
		self.readConfig()

	def chooseNVDAObjectOverlayClasses(self, obj, clsList):
		if obj.windowClassName == "TSStatusBarClass":# and obj.role == controlTypes.ROLE_STATICTEXT :
			clsList.insert(0, CheckStatusbarText)
		elif obj.windowClassName == 'RICHEDIT50W' and obj.role == controlTypes.ROLE_EDITABLETEXT and controlTypes.STATE_READONLY in obj.states and obj.windowControlID in [1006, 3011]  and sys.getwindowsversion().major == 10:
			clsList.insert(0, CheckHistoryChange)
		elif obj.windowControlID in [1393, 1396, 1399] and obj.role == controlTypes.ROLE_RADIOBUTTON:
			clsList.insert(0, RadioButtonsFindContact)

	def event_NVDAObject_init(self,obj):
		gw = winUser.getWindow
		gc = winUser.getClassName
		ga = winUser.getAncestor
		for x in xrange (65,91):
			self.bindGesture("kb:%s" %chr(x), "findContactInList")
			self.bindGesture("kb:%s" %chr(x+32), "findContactInList")
		if isinstance(obj, NVDAObjects.IAccessible.IAccessible) and gc(gw(ga(obj.windowHandle,2),5)) == "TSTabCtrlClass":
			if obj.windowControlID in [1002, 3012] and obj.role == controlTypes.ROLE_EDITABLETEXT: 
				# Translators: Name into the edit field of the messages.
				obj.name = _('Type a message here')
			elif obj.windowControlID in  [1006, 3011]:# and obj.role == controlTypes.ROLE_EDITABLETEXT and controlTypes.STATE_READONLY in obj.states:
			#elif gc == 'RICHEDIT50W':
				# Translators: Name for the object of history, the field read-only.
				obj.name = _('History')
		elif isinstance(obj, NVDAObjects.IAccessible.IAccessible) and obj.windowClassName == u'Miranda':
			# Translators: Name to the main window of Miranda NG
			obj.name = _('Miranda NG main window')
		elif isinstance(obj, NVDAObjects.IAccessible.IAccessible) and obj.role == controlTypes.ROLE_DIALOG:
			if obj.firstChild.windowClassName is not None and obj.firstChild.windowClassName == 'MHeaderbarCtrl': # Window of request of friendship.
				name = obj.firstChild.name
				pattern = re.compile('\\n\\d+ ')
				if pattern.search (name):
					name = pattern.sub (' ', name)
				#obj.name += obj.firstChild.name
				obj.name = '%s - %s' %(obj.name, name)
		elif isinstance(obj, NVDAObjects.IAccessible.IAccessible) and obj.role == controlTypes.ROLE_EDITABLETEXT and obj.windowControlID == 1394:
			obj.name = _("Facebook ID")
		elif isinstance(obj, NVDAObjects.IAccessible.IAccessible) and obj.role == controlTypes.ROLE_EDITABLETEXT and obj.windowControlID == 1048:
			obj.name = _("E-mail address")
	def readConfig(self):
		#Read configuration set from Miranda.ini
		self.announceMessageFromAnotherConversationTab, self.typingAnnounce, self.typingSounder, self.announceDateAndTime = 1, 1, 1, 1
		config_miranda = os.path.join (os.path.dirname(__file__),'Miranda.ini') 
		if os.path.isfile(config_miranda):
			config = ConfigObj(config_miranda, encoding = "utf-8")
			try:
				self.announceMessageFromAnotherConversationTab = eval(config['Miranda Add-on Settings']['Announce messages from another chat tab'])
				self.typingAnnounce = eval(config['Miranda Add-on Settings']['Announce who is typing'])
				self.typingSounder = eval(config['Miranda Add-on Settings']['Typing sound'])
				self.announceDateAndTime = eval(config['Miranda Add-on Settings']['Announce date and time in messages'])
			except (KeyError, IOError):
				pass 

	def playTicks (self):
	# Play a sound when someone is typing in the current chat window.
		msgIsTyping = _("is typing a message")
		#self.waveFile = os.path.join(os.path.dirname(__file__) + '\\Sounds\\MirandaTyping%d.wav' %self.typingSounder)
		waveFile = os.path.join(os.path.dirname(__file__), 'Sounds', 'MirandaTyping%d.wav' %self.typingSounder).decode("mbcs")
		if not os.path.isfile (waveFile):
			return
		fore = api.getForegroundObject()
		h = self.findDescendantWindow (api.getForegroundObject().windowHandle, 'TSStatusBarClass')
		if h:
			obj = NVDAObjects.IAccessible.getNVDAObjectFromEvent (h, -4, 0)
			statusbarText = obj.firstChild.name
			try:
				while msgIsTyping in statusbarText and fore.windowHandle == api.getForegroundObject().windowHandle:
					statusbarText = obj.firstChild.name
					winsound.PlaySound (waveFile,0)#winsound.SND_PURGE|winsound.    SND_ASYNC)
			except:
				pass

	def script_readMessage (self, gesture):
		#Read messages with the shortcut keys alt + numbers.
		# If pressed twice, displays the history in a dialog with its message in focus.
		t = self.getActiveChat ()
		if not t[0]: #
			# Translators: Message that informs you that the focus is not in a chat window.
			ui.message (_('Not in conversation window'))
			return
		h = t[1] # history window.
		key=gesture.mainKeyName
		n =   int(key) 
		#reindex 0 for location
		if n == 0:
			n=10
		# get history content.
		historyText = self.getHistory (h)
		if historyText == None:
			# Translators: Message for when the history is empty.
			ui.message (_('The history is empty'))
			return
		messagesList = self.adjustHistory (historyText)
		totalMsgs = len (messagesList)
		if n > totalMsgs:
			ui.message (_("There's so many messages"))
			return
		position = totalMsgs - n
		msg = messagesList [position]
		ui.message (msg)
		self.lastMessage = msg
		# saves the index of the last message read.
		self.dictPositions[h] = position
		if scriptHandler.getLastScriptRepeatCount():
			self.showHistory (messagesList, position)
	script_readMessage.__doc__ = _("Read messages when you press alt plus numbers, 1 being the most recent")

	def showHistory (self, messagesList, position):
		# Prepare the display of the selected message in the history.
		aux = '\r\n'.join (messagesList)#[:position]) 
		aux1 = '\r\n'.join(messagesList[:position])
		if not self.announceDateAndTime: # no date and time.
			#pat = re.compile ('  \d\d/\d\d/\d\d\d\d| Today| - \d\d:\d\d')
			pat = re.compile ('  \d\d/\d\d/\d\d\d\d| Today| Yesterday| - \d\d:\d\d')
			if pat.search (aux):
				aux = pat.sub('', aux)
			if pat.search (aux1):
				aux1 = pat.sub('', aux1)
		index =len(aux1)+2 * (position != 0)
		contactName = api.getForegroundObject().name.split(' (')[0]
		title = _('Miranda Add-on - Chat history with %s. Total of messages: %d') %(contactName, len(messagesList))
		# Translators: The final message contained in the dialog history view to close it.
		self.showStaticTextDialog (title, _('%s%sPress Escape to close this dialog.') %(aux ,'\r\n\r\n'), index)


	def getHistory (self, h):
		# gets the contents of the history
		historyText = None
		obj = NVDAObjects.IAccessible.getNVDAObjectFromEvent (h, 0-4, 0)
		for child in obj.children:
			if controlTypes.STATE_READONLY  in child.states and child.role == controlTypes.ROLE_EDITABLETEXT:
				hwnd = child.windowHandle
				break
		buf_size = 1 + winUser.sendMessage(hwnd, win32con.WM_GETTEXTLENGTH, 0, 0)
		buffer =ctypes.create_unicode_buffer(buf_size) 
		winUser.sendMessage(hwnd, win32con.WM_GETTEXT, buf_size, buffer)
		historyText = buffer[:buf_size-1].strip()    
		if historyText == '':
			return None
		historyText = historyText.replace ('\r\n\r\n', '\r\n')
		pattern = re.compile (' ~-\+\d+\+-~')
		if pattern.search (historyText):
			historyText = pattern.sub ('', historyText)
		return historyText

	def adjustHistory (self, historyText, lastMessage = False):
		# organizes the history.
		ret = []
		l = historyText.split ('\xad\r\n'.decode('latin-1'))
		if lastMessage:
			l = [l[-1]]
		for c in xrange (len(l)):
			text = l[c]
			name = text.split('\r\n')[0]
			pat = re.compile ('\d \d\d:\d\d')
			h = pat.findall(text)
			m = pat.split(text)
			total = len(h)
			for u in xrange(len(h)):
				msg = '%s - %s:\r\n%s' %(name, h[u][2:], m[u+1])
				if not self.announceDateAndTime: # no date and time.
					pat = re.compile ('  \d\d/\d\d/\d\d\d\d| Today| Yesterday| - \d\d:\d\d')
					if pat.search (msg):
						msg = pat.sub('', msg)
				ret.append (msg)
		return ret

	def script_relativeMsg (self, gesture):
		# Enables you to read all the messages contained in the history.
		# The last read message also can be copied to the Clipboard.
		h = self.getActiveChat ()[1]
		if not h:
			gesture.send()
			return
		historyText = self.getHistory (h)
		if historyText is None:
			# Translators: Message that indicates that the history is empty.
			ui.message (_('The history is empty'))
			return
		messagesList = self.adjustHistory (historyText)
		totalMsgs = len (messagesList)  - 1
		# Retrieves the index of the last message read.
		n = self.dictPositions [h]
		if n < 0:
			newIndex = totalMsgs
		if gesture.mainKeyName == "enter":
			self.showHistory (messagesList, n)
			return
		elif gesture.mainKeyName == ".":
			msg = messagesList[n]
			api.copyToClip (msg)
			# Translators: Message that indicates that the last read message has been sent to the clipboard.
			ui.message (_('the last read message was sent to the clipboard'))
			return
		elif gesture.mainKeyName == 'upArrow' and n > 0:
			newIndex= n - 1
		elif gesture.mainKeyName == 'downArrow' and n < totalMsgs:
			newIndex =n + 1
		elif gesture.mainKeyName == 'home':
			newIndex = 0
		elif gesture.mainKeyName == 'end':
			newIndex = totalMsgs
		elif gesture.mainKeyName == 'pageUp':
			newIndex = (n-10) * (n>=10)
		elif gesture.mainKeyName == 'pageDown':
			if totalMsgs - n >= 10:
				newIndex = n + 10
			else:
				newIndex = totalMsgs
		else:
			newIndex = n
		self.dictPositions [h] = newIndex
		msg = messagesList[newIndex]
		ui.message (msg)
		self.lastMessage = msg
		if newIndex == 0 or newIndex == totalMsgs:
			# first or last message.
			tones.beep (1500,10)
	script_relativeMsg.__doc__ = _("Read messages from history in chat window: Alt + down arrow, to the next message; ALT + up arrow, to the previous message; ALT + pageDown, 10 messages below; ALT + PageUP, 10 messages above; ALT + home, first message; ALT + End, last message; ALT + period, copies the last message read to the Clipboard.")

	def script_list_Links_emails (self, gesture):
		# Gets the links or emails contained in the history.
		# The links can be opened in the default browser or copied to the Clipboard.
		# The emails can be opened in the default email client or copied to the Clipboard.
		h = self.getActiveChat ()[1]
		if not h:
			# Translators: Message that indicates that the focus is not in a chat window.
			ui.message (_('Not in conversation window'))
			return
		historyText = self.getHistory (h)
		if historyText is None:
			# Translators: Message that indicates that the history is empty.
			ui.message (_('The history is empty'))			
			return
		key=gesture.mainKeyName
		op= (key!='f7')
		# op = 0, search for links.
		# op = 1, search for emails.
		pat= (re.compile (r'((\bhttp[s]?://|\bwww\.|\bftp[s]?://)[-a-zA-z0-9+&@#/%?=~_|!:,.;]*[-a-zA-Z0-9+&@#/%=~_|])'), re.compile('([\w\-\.]+@(\w[\w\-]+\.)+[\w\-]+)'))[op]
		if pat.search (historyText) is None:
			ui.message([
			# Translators: Message that indicates that there is no links in the history:
			_('There are no links in history'),
			# Translators: Message that indicates that there are no emails in the history.
			_('There are no E-mails in history')
			][op])
			return
		l = pat.findall (historyText)
		resultsList = list(set([x[0] for x in l]))
		resultsList.reverse() 
		if not op: # links
			# Translators: The title of the dialog box that lists the links contained in the history.
			title=_("Links list")
			# Translators: Label of the dialog box that lists the links contained in the history.
			label=_('Select a link to open or  to copy to clipboard')
		else: # e-mails
			# Translators: The title of the dialog box that lists the e-mails contained in the history.
			title=_("E-mails list")
			# Translators: Label of the dialog box that lists the e-mails  contained in the history.
			label=_('Select a e-mail to send a message or  to copy to clipboard')
		self.showDialog(0, resultsList, title, label)
	script_list_Links_emails.__doc__ = _("Displays a dialog box with the links or emails contained in the history of a chat window.")

	def event_foreground (self, obj, nextHandler):
		if appModuleHandler.getAppModuleForNVDAObject(api.getFocusObject()) == appModuleHandler.getAppModuleForNVDAObject(obj):
			h = self.getActiveChat ()[1]
			if  h and api.getFocusObject().role == controlTypes.ROLE_DIALOG:
			# Identifies the chat window.
				ui.message (_('Miranda NG Chat'))
				self.adjustInitialPosition ()
				return
			elif self.isInFindContactWindow():
				x=0
				for c in [1393, 1396, 1399]:
					h = self.findDescendantWindow (obj.windowHandle, c)
					rb=NVDAObjects.IAccessible.getNVDAObjectFromEvent(h, -4, 0)
					if controlTypes.STATE_CHECKED in rb.states:
						self.filterFindContact = x
						break
					x+=1
		nextHandler()


	def event_selection (self, obj, nextHandler):
		if self.isInContactList () and api.getForegroundObject().role == controlTypes.ROLE_PANE and obj.role == controlTypes.ROLE_LISTITEM:
			if self.lastFocus == obj.name:
				return
		elif obj.windowClassName == 'TSTabCtrlClass' and obj.role == controlTypes.ROLE_TAB:
			speech.speakObjectProperties (obj, name=True, role=True)
			self.lastMessage = None
			self.adjustInitialPosition ()
		nextHandler()

	def adjustInitialPosition (self):
		# Assigns a initial position in the dictionary for a chat window open or remove if it does not exist.
		h = self.getActiveChat ()[1]
		keyExists = False
		for  key in self.dictPositions.keys():
			if not winUser.isWindowEnabled (key):
				del (self.dictPositions[key])
			elif h == key:
				keyExists = True
		if not keyExists: 
				self.dictPositions[h] = -1
				self.lastMessage = None

	def event_focusEntered (self, obj, nextHandler):
		if obj.role == controlTypes.ROLE_PROPERTYPAGE:
			if sys.getwindowsversion().major < 10:
				return
			h = self.findDescendantWindow(obj.windowHandle, 1006)
			if not h:
				h	 = self.findDescendantWindow(obj.windowHandle, 3011)
			#obj = obj.firstChild.simpleNext.simpleNext.simpleNext
			obj = NVDAObjects.IAccessible.getNVDAObjectFromEvent (h, -4, 0)
			if self.chatWindow is not None:
				self.chatWindow.stopMonitoring ()
			self.chatWindow = obj
			self.chatWindow.startMonitoring()
			return
		nextHandler()


	def event_gainFocus (self, obj, nextHandler):
		if self.isInContactList():
			if self.lastFocus == obj.name:
				return
		self.lastFocus = obj.name
		h = self.findDescendantWindow (api.getForegroundObject().windowHandle, 'TSStatusBarClass')
		if h:
			if self.statusbar is None:
				self.statusbar = NVDAObjects.IAccessible.getNVDAObjectFromEvent (h, -4, 0)
				self.statusbar.startMonitoring()
		else :
			if self.statusbar is not None:
				self.statusbar.stopMonitoring ()
				self.statusbar = None
		nextHandler()

	def event_appModule_loseFocus(self):
		if self.statusbar is not None:
			try:
				self.statusbar.stopMonitoring()

			except:
				pass
			self.statusbar = None

		if self.chatWindow is not None:
			self.chatWindow.stopMonitoring()
			self.chatWindow = None
		if  self.wmp is not None and self.wmp.playstate== 3: # the audio clip is being played.:
			self.wmp.controls.stop()

	def script_movesFocusToChatReceivedFromAnotherWindow (self, gesture):
		t = self.getActiveChat ()
		if not t[0]: #
			# Translators: Message that informs you that the focus is not in a chat window.
			ui.message (_('Not in conversation window'))
			return
		if sys.getwindowsversion().major == 10:
			# Translators: Message that informs that the resource is not available for windows 10.
			ui.message (_('Not available for windows 10'))
			return
		obj = api.getForegroundObject().firstChild
		if self.lastContactMessage is None or not self.announceMessageFromAnotherConversationTab:
			# Translators: The message that informs you that has not been received chat to another tab or the resource is not available.
			ui.message (_('Has not yet received any message from another chat tab or this resource is not available'))
			return
		obj = api.getForegroundObject().firstChild.firstChild.firstChild
		while obj is not None:
			if self.lastContactMessage in obj.name:
				clickObject (obj.location)
				return
			obj = obj.next
		if obj is None:
			# Translators: Message that informs that there was an error in the transition to another tab of chat.
			ui.message (_('Error trying to open the chat'))
	script_movesFocusToChatReceivedFromAnotherWindow.__doc__ = _("Moves the focus to the last tab where a message was received.")

	def getActiveChat (self):
		# Returns a tuple with the number of active conversations and the identifier of the current history.
		h=winUser.getForegroundWindow()
		h = winUser.getWindow (h,5)
		if winUser.getClassName(h) != 'TSTabCtrlClass':
			return (0,0)
		h = winUser.getWindow (h,5)
		j = 0
		x = 0
		while h:
			if winUser.isWindowVisible (h): # active tab.
				j = h
			h = winUser.getWindow (h,2)
			x += 1
		return (x, j)

	def event_stateChange (self, obj, nextHandler):
		controlsID = [1393, 1396, 1399]
		if obj.windowControlID in  controlsID and controlTypes.STATE_CHECKED in obj.states:
			self.filterFindContact = controlsID.index(obj.windowControlID)
		nextHandler()


	def event_valueChange(self, obj, nextHandler):
		t = self.getActiveChat()	
		if t[0] and obj.role == controlTypes.ROLE_EDITABLETEXT and controlTypes.STATE_READONLY in obj.states and api.getForegroundObject().firstChild.windowClassName == "TSTabCtrlClass" and sys.getwindowsversion().major < 10: 
			# getting the history of the active chat window.
			historyText = self.getHistory (obj.parent.parent.windowHandle)
			if historyText is None:
				return
			# Organizing the text of the history.
			l = self.adjustHistory (historyText, True)
			if l==[]:
				return
			value = l [-1]
			if self.announceDateAndTime:
				contactSeparator = '  '
			else:
				contactSeparator = ':'
				pat = re.compile ('  \d\d/\d\d/\d\d\d\d| Today| Yesterday| - \d\d:\d\d')
				if pat.search (value):
					value = pat.sub('', value)
			contactName = value.split (contactSeparator)[0]
			if winUser.isWindowVisible (obj.windowHandle): # actual chat.
				if value != self.lastMessage:
					ui.message (value)
					self.lastMessage = value
					if self.lastContactMessage != contactName:
						self.lastContactMessage = None
			else: # Message from another tab.
				if self.lastContactMessage != contactName and self.announceMessageFromAnotherConversationTab:
					ui.message ('mensagem de %s' %contactName)
					self.lastContactMessage  = contactName
		nextHandler()

	def isInContactList (self):
		#  Returns true if focus is in the Miranda main window.
		obj = api.getFocusObject()
		fore = api.getForegroundObject()
		return (winUser.getClassName(fore.windowHandle) == u'Miranda' and fore.role == controlTypes.ROLE_PANE and obj.role == controlTypes.ROLE_LISTITEM)

	def isInFindContactWindow (self):
		obj = api.getForegroundObject()
			# Translators: Title of Find/add Contact window. See Miranda translation.
		return obj.name == _('Find/add contacts')

	def script_changeTab (self, gesture):
		if self.isInContactList():
			# Searches for an active chat window, if any, moves the focus to it.
			obj = api.getForegroundObject().simpleParent.simpleNext
			while obj is not None and  appModuleHandler.getAppModuleForNVDAObject(obj).appName.upper() != 'MIRANDA64' and appModuleHandler.getAppModuleForNVDAObject(obj).appName.upper() != 'MIRANDA32':
				obj = obj.simpleNext
			if obj is not None:
				obj.firstChild.next.setFocus()
			else:
			# Translators: Message that informs there are no chats open
				ui.message (_('There is no open chat window'))
			return
		gesture.send()
		t = self.getActiveChat ()
		nChats = t[0]
		if nChats == 1:
			# Translators: Message that informs that there is only one chat open.
			ui.message (_('This is the only active chat window'))

	def findDescendantWindow (self, h, c):
		# Returns the handle of a window with a given class sought from a specified window.
		try:
			if type (c) is int:
				return windowUtils.findDescendantWindow (h, controlID=c)
			else:
				return windowUtils.findDescendantWindow (h, className=c)
		except LookupError:
			return False

	def script_configAddon (self, gesture):
		# Displays the dialog box to configure the add-on.
		# Translators: Title of the configuration window .
		title = _('Miranda add-on settings')
		# Translators: the label options dialog box settings.
		label = _('Options')
		self.readConfig()
		# Translators: The list of configuration options.
		list = [[_('Announce messages from another chat tab'), self.announceMessageFromAnotherConversationTab], [_('Announce who is typing'), self.typingAnnounce], [_('Typing sound'), self.typingSounder], [_('Announce date and time in messages'), self.announceDateAndTime]]
		self.showDialog (1, list, title, label)
	script_configAddon.__doc__ = _("Displays the dialog box to configure the add-on.")

	def showDialog (self, op, list, title, label,selectItem=0):
		# displays a list box with a certain content.
		# op=0, links or e-mails list.
		# op = 1, List for the add-on configuration.
		self._listDialog=ListDialog (gui.mainFrame,list, title,label, op)
		self._listDialog.updateList(list,selectItem)
		if not self._listDialog.IsShown():
			gui.mainFrame.prePopup()
			self._listDialog.Centre()
			self._listDialog.Show()
			gui.mainFrame.postPopup()

	def showStaticTextDialog (self, title, text, index):
		# Opens a dialog box with a given message.
		if self._staticTextDialog is None or True:
			self._staticTextDialog=StaticTextDialog(gui.mainFrame, title, text, index)
		else:
			self._staticTextDialog.title = title
			self._staticTextDialog.text = text + 'marcos'
			self._staticTextDialog.index = index
		if not self._staticTextDialog.IsShown():
			gui.mainFrame.prePopup()
			self._staticTextDialog.Centre()
			self._staticTextDialog.Show()
			gui.mainFrame.postPopup()

	def saveConfig (self, announceMessageFromAnotherConversationTab, typingAnnounce, typingSounder, announceDateAndTime):
		# saves changes settings.
		config = ConfigObj()
		config_miranda = os.path.join(os.path.dirname(__file__),'Miranda.ini') 
		config.filename = config_miranda
		config ["Miranda Add-on Settings"] = {"Announce messages from another chat tab": announceMessageFromAnotherConversationTab, "Announce who is typing": typingAnnounce, "Typing sound": typingSounder, "Announce date and time in messages":announceDateAndTime}
		try:
			config.write()
		except IOError:
			pass

	def event_appModule_gainFocus(self):
		self.readConfig ()

	def script_sayTotalChats (self, gesture):
		# Announces the existing chats.
		h = self.getActiveChat ()[0]
		if not h:
			# Translators: Message that informs you that the focus is not in a chat window.
			ui.message (_('Not in conversation window'))
			return
		obj = api.getForegroundObject().firstChild.firstChild
		l = []
		for child in obj.children:
			if child.role == controlTypes.ROLE_TAB:
				l.append (child.name.split (' (')[0])
			else:
			 break
		#Translators: Message that announces the total number of chats open.
		msg = _('Number of active chats: %d.%s%s') %(self.getActiveChat ()[0], '\r\n', '.\r\n'.join(l))
		ui.message (msg)
	script_sayTotalChats.__doc__ = _("When in a chat window, announces the total of open tabs and their respective names.")

	def script_playAudioClip (self, gesture):
		t = self.getActiveChat ()
		if not t[0]: #
			# Translators: Message announcing that the focus is not in a chat window.
			ui.message (_('Not in conversation window'))
			return
		h = t[1] # handle history window.
		historyText =self.getHistory(h)
		if historyText == None:
			# Message that informs that the history is empty.
			ui.message (_('The history is empty'))
			return
		# Getting the last read message in the current chat window.
		messagesList = self.adjustHistory (historyText)
		n = self.dictPositions [h]
		msg = messagesList[n]
		# find audio clip in the message.
		pat = re.compile ('<audioclip-\d+-\d+\....> (.+=1)')
		if pat.search(msg) is None:
			# Translators: The message that informs that was not found audio clip in the last message read.
			ui.message (_("Was not found audio clip in this message"))
			return
		audioClip = pat.findall (msg)[0]
		# Creating wmp COM object
		if self.wmp is None:
			import comtypes
			import comtypes.client as cc
			self.wmp = cc.CreateObject("WMPlayer.OCX")
		if self.wmp.playstate== 3: # the audio clip is being played.
			self.wmp.controls.stop()
			return
		self.wmp.url = audioClip
		self.wmp.controls.play()
		# Translators: Message to waitt the playback of the audio clip.
		ui.message (_('Please, wait...'))
		time.sleep (1.5)
		# checking wmp status
		th=Thread (target = self.checkWmpState, args = ())
		th.start()
	script_playAudioClip.__doc__ = _("Plays or stops an audio clip contained in the message ")

	def checkWmpState (self):
		x = 1
		while (x <= 2 and self.wmp.playstate==9) or self.wmp.playstate == 3:
			time.sleep (1)
			x+=1
		if self.wmp.playstate == 1: # audio clip stopped
			#Translators: Message of  audio clip stopped.
			ui.message (_('audio clip stopped'))
		elif self.wmp.playstate == 10: # wmp ready.
			# Translators: Message that indicates that the link is expired.
			ui.message (_('link expired'))

	def script_findContactInList (self, gesture):
		obj = api.getFocusObject()
		if not (self.isInContactList () and obj.windowClassName == 'ListBox' and obj.role == controlTypes.ROLE_LISTITEM):
			gesture.send()
			return
		key=gesture.mainKeyName
		total = obj.simpleParent.childCount
		if total==1:
			return
		startItem = obj.IAccessibleChildID
		if startItem == total:
			obj = obj.parent.firstChild
		else:
			obj = obj.next
		import unicodedata
		x = 0
		while obj is not None:
			currentItem = obj.IAccessibleChildID
			firstLetter = obj.name.strip()[0].upper()
			nfkd = unicodedata.normalize('NFKD', firstLetter)
			f = u"".join([c for c in nfkd if not unicodedata.combining(c)])
			if str(key).upper() in f and currentItem != startItem and currentItem!= 1:
				obj.setFocus()
				clickObject(obj.location)
				break
			elif currentItem == startItem and x == total-1:
				tones.beep (500,30)
				break
			elif currentItem == total:
				obj = obj.parent.firstChild
			else:
				obj = obj.next
			x+=1


	def script_tab (self, gesture):
		if self.isInFindContactWindow ():
			obj = api.getFocusObject()
			id = obj.windowControlID
			if self.filterFindContact == 0:
				if id == 1394:
					obj = obj.simpleNext.simpleNext.simpleNext.simpleNext.simpleNext.simpleNext.simpleNext.simpleNext.simpleNext
					obj.setFocus()
			elif self.filterFindContact == 1:
				if id == 1245:
					obj=obj.simpleNext.simpleNext			
					obj.setFocus()
				elif id == 1048:
					obj = obj.simpleNext.simpleNext.simpleNext.simpleNext.simpleNext.simpleNext.simpleNext
					obj.setFocus()
			elif self.filterFindContact == 2:
				if id== 1245:
					obj = obj.simpleNext.simpleNext.simpleNext.simpleNext
					obj.setFocus()
		gesture.send()

	def script_shiftTab (self, gesture):
		if self.isInFindContactWindow():
			obj = api.getFocusObject()
			id = obj.windowControlID
			if self.filterFindContact == 0:
				if id== 1:
					obj=obj.simplePrevious.simplePrevious.simplePrevious.simplePrevious.simplePrevious.simplePrevious.simplePrevious.simplePrevious.simplePrevious
					obj.setFocus()
			elif self.filterFindContact == 1:
				if id== 1:
					obj=obj.simplePrevious.simplePrevious.simplePrevious.simplePrevious.simplePrevious.simplePrevious.simplePrevious
					obj.setFocus()
				elif id == 1396:
					obj = obj.simplePrevious.simplePrevious
					obj.setFocus()
			elif self.filterFindContact == 2:
				if id== 1399:
					obj=obj.simplePrevious.simplePrevious.simplePrevious
					obj.setFocus()
		gesture.send()

	
	__gestures ={
		"kb:control+t": "sayTotalChats",
		"kb:control+k": "movesFocusToChatReceivedFromAnotherWindow",
		"kb:f2": "playAudioClip",
		"kb:f7": "list_Links_emails",
		"kb:f9": "list_Links_emails",
		"kb:alt+upArrow": "relativeMsg",
		"kb:alt+downArrow": "relativeMsg",
		"kb:alt+home": "relativeMsg",
		"kb:alt+end": "relativeMsg",
		"kb:Alt+Enter": "relativeMsg",
		"kb:control+pageUp": "relativeMsg",
		"kb:control+pageDown": "relativeMsg",
		"kb:alt+.": "relativeMsg",
		"kb:alt+1": "readMessage",
		"kb:alt+2": "readMessage",
		"kb:alt+3": "readMessage",
		"kb:alt+4": "readMessage",
		"kb:alt+5": "readMessage",
		"kb:alt+6": "readMessage",
		"kb:alt+7": "readMessage",
		"kb:alt+8": "readMessage",
		"kb:alt+9": "readMessage",
		"kb:alt+0": "readMessage",
		"kb:control+tab": "changeTab",
		"kb:control+shift+tab": "changeTab",
		"kb:alt+c": "configAddon",
		"kb:tab": "tab",
		"kb:shift+tab": "shiftTab",
		}

class ListDialog(wx.Dialog):
	def __init__(self, parent, list, title, labelCaption, typeList):
		self.list = list
		self.typeList=typeList
		self.states=[]
		if self.typeList==1: #check list.
			self.states=([obj[1] for obj in self.list])
		super(ListDialog, self).__init__(parent, title=title)
		self.title=title
		if self.typeList==0: # lista de e-mails ou links.
			labelButton1=[
			# Translators: button to send e-mail with the e-mail client default.
			_('&Send e-mail with default client')
			,
			# Translators: button to open the selected link with the default browser.
			_('&Open link with default browser')
			][not '@' in self.list[0]]

			labelButton2=[
			# Translators: button to copy the selected email to the clipboard.
_('&Copy e-mail to clipboard')
			,
			# Translators: Button to copy the selected link to the clipboard.
			_('&Copy link to clipboard')
			][not '@' in self.list[0]]
		else:
			labelButton1='Ok'
		# create interface.
		mainSizer = wx.BoxSizer(wx.VERTICAL)
		tasksSizer = wx.BoxSizer(wx.VERTICAL)
		# create listview.
		tasksLabel = wx.StaticText(self, -1, label=labelCaption)
		tasksSizer.Add(tasksLabel)
		self.listBox = wx.ListCtrl(self, size=(550, 250),style=wx.LC_REPORT|wx.BORDER_SUNKEN)
		tasksSizer.Add(self.listBox, proportion=8)
		#mainSizer.Add(tasksSizer)
		# create buttons.
		buttonsSizer = wx.BoxSizer(wx.HORIZONTAL)
		button1ID=wx.NewId()
		self.button1 = wx.Button(self,button1ID, labelButton1)
		buttonsSizer.Add(self.button1)
		if self.typeList==0: # lista de e-mails ou links , exibe mais 1 botão.
			button2ID=wx.NewId()
			self.button2 = wx.Button(self, button2ID, labelButton2)
			buttonsSizer.Add(self.button2)
			# Defines the events for the second button.
			wx.EVT_KILL_FOCUS (self.button2,self.onLostFocus)
			self.Bind(wx.EVT_BUTTON,self.onButton2,id=button2ID)
		self.cancelButton = wx.Button(self, wx.ID_CANCEL)
		buttonsSizer.Add(self.cancelButton)
		tasksSizer.Add(buttonsSizer)
		mainSizer.Add(tasksSizer)
		# Defines the events for other controls.
		wx.EVT_KILL_FOCUS (self.listBox,self.onLostFocus)
		wx.EVT_KILL_FOCUS (self.button1,self.onLostFocus)
		wx.EVT_KILL_FOCUS (self.cancelButton,self.onLostFocus)
		self.Bind(wx.EVT_BUTTON,self.onButton1,id=button1ID)
		self.listBox.Bind(wx.EVT_KEY_DOWN, self.onKeyPress)
		mainSizer.Fit(self)
		self.SetSizer(mainSizer)


	def onButton1 (self,evt):
		self.Hide()
		if not self.typeList:
			self.openLink()
		else: # configurations, check list.
			self.checkChanges()
		evt.Skip()

	def onButton2 (self,evt):
		self.Hide()
		index=self.listBox.GetFocusedItem()
		if not self.typeList: # lista de links ou e-mails.
			item=self.listBox.GetItemText(index)
			api.copyToClip(item)
		evt.Skip()

	def onKeyPress(self, event):
		keycode = event.GetKeyCode()
		if keycode==wx.WXK_RETURN:
			self.Hide()
			if self.typeList==0: # lista de links ou e-mails.
				self.openLink()

		elif  keycode == wx.WXK_SPACE and self.listBox.GetColumnCount()==2: # check list.
			# check or uncheck the items from the list.
			index=self.listBox.GetFocusedItem()
			if index != 2:
				newStatus=int(not self.list[index][1])
			else:
				newStatus=int(self.list[index][1]) + 1 - 5 * (int(self.list[index][1]) == 4)
			self.list[index][1]=newStatus
			self.updateList(self.list,index)
		event.Skip()

	def openLink (self):
		# Open a link with the default browser or open the default mail client to send a message.
		import os
		index=self.listBox.GetFocusedItem()
		item=self.listBox.GetItemText(index)
		if '@' in item:
			os.startfile('mailto:%s' %item)
		else:
			os.startfile(item)

	def checkChanges (self):
		# If there have been changes, saves the changes in the ini file.
		currentStatesList = [x[1] for x in self.list]
		if self.states != currentStatesList: # houve alterações.
			AppModule(None).saveConfig (currentStatesList[0], currentStatesList[1], currentStatesList[2], currentStatesList[3])

	def onLostFocus(self,evt):
		obj = api.getForegroundObject()
		hasFocus= False
		for child in obj.children:
			hasFocus+= child.hasFocus
		if not hasFocus:
			self.Hide()
			evt.Skip()

	def updateList(self, list, index):
		self.listBox.ClearAll()
		self.listBox.InsertColumn(0,_('Item'))
		if not self.typeList: # não é um check list.
			for item in list:
				self.listBox.Append ([item])
		else: # check list.
			self.listBox.InsertColumn(1, _('Status'))
			op=[_("Disabled"),_("Enabled")]
			op1=[_("Disabled"),_("Volume 1"), _('Volume 2'), _('Volume 3'), _('Volume 4')]
			c=1
			for x in list:
				if c != 3:
					self.listBox.Append ([x[0],op[x[1]]])
				else:
					self.listBox.Append ([x[0],op1[x[1]]])
				c += 1
		self.listBox.Focus(index)
		self.listBox.Select(index)

class StaticTextDialog(wx.Dialog):
	def __init__(self, parent, title, text, index):
		self.text = text
		self.title=title
		self.index = index
		super(StaticTextDialog, self).__init__(parent, title=title)
		# Create interface
		mainSizer = wx.BoxSizer(wx.VERTICAL)
		tasksSizer = wx.BoxSizer(wx.VERTICAL)
		# Create a static text.
		read_only_txtCtrlId =120#wx.NewId()
		self.read_only_txtCtrl = wx.TextCtrl(self,read_only_txtCtrlId,self.text,style=wx.TE_MULTILINE|wx.TE_READONLY, size=(800,200))
		tasksSizer.Add(self.read_only_txtCtrl)
		mainSizer.Add(tasksSizer)
		# Create Close button.
		buttonsSizer = wx.BoxSizer(wx.HORIZONTAL)
		self.closeButton = wx.Button(self, wx.ID_CANCEL, _('&Close'))
		buttonsSizer.Add(self.closeButton)
		# TextCtrl and buttons events.
		wx.EVT_KILL_FOCUS (self.read_only_txtCtrl,self.onLostFocus)
		wx.EVT_KILL_FOCUS (self.closeButton,self.onLostFocus)
		wx.EVT_CHAR (self.read_only_txtCtrl,self.onKeyPress)

		tasksSizer.Add(buttonsSizer)
		mainSizer.Add(tasksSizer)
		mainSizer.Fit(self)
		self.SetSizer(mainSizer)
		self.read_only_txtCtrl.ShowPosition (index)
		self.read_only_txtCtrl.SetInsertionPoint(index)
		#self.read_only_txtCtrl.SetFocus()



	def onKeyPress(self, event):
		keycode = event.GetKeyCode()
		# verifica se ctrl+A foi pressionado para copiar todo o texto.
		if keycode==1:
			self.read_only_txtCtrl.SelectAll()
		event.Skip()

	def onLostFocus(self,evt):
		obj = api.getForegroundObject()
		hasFocus= False
		for child in obj.children:
			hasFocus+= child.hasFocus
		if not hasFocus:
			self.Hide()

			evt.Skip()
